@@ -1,21 +1,21 @@
 /* pocl/_kernel_c.h - C compatible OpenCL types and runtime library
-   functions declarations.
+   functions declarations for kernel builtin implementations using C.
 
    Copyright (c) 2011 Universidad Rey Juan Carlos
-   Copyright (c) 2011-2013 Pekka Jääskeläinen / TUT
+   Copyright (c) 2011-2017 Pekka Jääskeläinen / TUT
    Copyright (c) 2011-2013 Erik Schnetter <eschnetter@perimeterinstitute.ca>
                            Perimeter Institute for Theoretical Physics
-   
+
    Permission is hereby granted, free of charge, to any person obtaining a copy
    of this software and associated documentation files (the "Software"), to deal
    in the Software without restriction, including without limitation the rights
    to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
    copies of the Software, and to permit persons to whom the Software is
    furnished to do so, subject to the following conditions:
-   
+
    The above copyright notice and this permission notice shall be included in
    all copies or substantial portions of the Software.
-   
+
    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
@@ -33,6 +33,64 @@
 
 #include "pocl_types.h"
 
+#if (__clang_major__ == 3)
+# if (__clang_minor__ == 7)
+# undef LLVM_3_7
+# define LLVM_3_7
+#elif (__clang_minor__ == 8)
+# undef LLVM_3_8
+# define LLVM_3_8
+#elif (__clang_minor__ == 9)
+# undef LLVM_3_9
+# define LLVM_3_9
+#endif
+
+#elif (__clang_major__ == 4)
+
+# undef LLVM_4_0
+# define LLVM_4_0
+
+#elif (__clang_major__ == 5)
+
+# undef LLVM_5_0
+# define LLVM_5_0
+
+#else
+
+#error Unsupported Clang/LLVM version.
+
+#endif
+
+#if (defined LLVM_3_6)
+# define LLVM_OLDER_THAN_3_7 1
+# define LLVM_OLDER_THAN_3_8 1
+# define LLVM_OLDER_THAN_3_9 1
+# define LLVM_OLDER_THAN_4_0 1
+# define LLVM_OLDER_THAN_5_0 1
+#endif
+
+#if (defined LLVM_3_7)
+# define LLVM_OLDER_THAN_3_8 1
+# define LLVM_OLDER_THAN_3_9 1
+# define LLVM_OLDER_THAN_4_0 1
+# define LLVM_OLDER_THAN_5_0 1
+#endif
+
+#if (defined LLVM_3_8)
+# define LLVM_OLDER_THAN_3_9 1
+# define LLVM_OLDER_THAN_4_0 1
+# define LLVM_OLDER_THAN_5_0 1
+#endif
+
+#if (defined LLVM_3_9)
+# define LLVM_OLDER_THAN_4_0 1
+# define LLVM_OLDER_THAN_5_0 1
+#endif
+
+#if (defined LLVM_4_0)
+# define LLVM_OLDER_THAN_5_0 1
+#endif
+
 #include "_kernel_constants.h"
 
 /* Function/type attributes supported by Clang/SPIR */
@@ -51,9 +109,7 @@
 #else
 #  define _CL_OVERLOADABLE
 #endif
-#if (__clang_major__ == 3) && (__clang_minor__ >= 2)
-/* This causes an error with Clang 3.1: */
-/* #if __has_attribute(__const__) */
+#if __has_attribute(__const__)
 #  define _CL_READNONE __attribute__((__const__))
 #else
 #  define _CL_READNONE
@@ -106,19 +162,17 @@ typedef uint uint8  __attribute__((__ext_vector_type__(8)));
 typedef uint uint16 __attribute__((__ext_vector_type__(16)));
 
 #if defined(__CBUILD__) && defined(cl_khr_fp16)
-/* NOTE: the Clang's __fp16 does not work robustly in C mode, 
+/* NOTE: the Clang's __fp16 does not work robustly in C mode,
    it might produce invalid code at least with half vectors.
    Using the native 'half' type in OpenCL C mode works better. */
 typedef __fp16 half;
 #endif
 
-#ifdef cl_khr_fp16
 typedef half half2  __attribute__((__ext_vector_type__(2)));
 typedef half half3  __attribute__((__ext_vector_type__(3)));
 typedef half half4  __attribute__((__ext_vector_type__(4)));
 typedef half half8  __attribute__((__ext_vector_type__(8)));
 typedef half half16 __attribute__((__ext_vector_type__(16)));
-#endif
 
 typedef float float2  __attribute__((__ext_vector_type__(2)));
 typedef float float3  __attribute__((__ext_vector_type__(3)));
@@ -158,7 +212,7 @@ typedef ulong ulong16 __attribute__((__ext_vector_type__(16)));
    the default builtins we use C functions which require
    the typedefs to the actual underlying types.
 */
-#if defined(__CBUILD__)
+#if defined(__CBUILD__) && defined(CLANG_OLDER_THAN_3_9)
 typedef int sampler_t;
 
 /* Since some built-ins have different return types
@@ -169,6 +223,7 @@ typedef int sampler_t;
  * dev_image_t. The structs are not anonymous to allow identification
  * by name.
  */
+
 typedef struct _pocl_image2d_t { dev_image_t base; }* image2d_t;
 typedef struct _pocl_image3d_t { dev_image_t base; }* image3d_t;
 typedef struct _pocl_image1d_t { dev_image_t base; }* image1d_t;
@@ -177,88 +232,35 @@ typedef struct _pocl_image2d_array_t { dev_image_t base; }* image2d_array_t;
 typedef struct _pocl_image1d_array_t { dev_image_t base; }* image1d_array_t;
 #endif
 
+#ifdef POCL_USE_FAKE_ADDR_SPACE_IDS
+/*
+ * During pocl kernel compiler transformations we use the fixed address
+ * space ids of clang's -ffake-address-space-map to mark the different
+ * address spaces to keep the processing target-independent. These
+ * are converted to the target's address space map (if any), in a final
+ * kernel compiler pass (TargetAddressSpaces). This is deprecated and
+ * will go after https://reviews.llvm.org/D26157 is available in the
+ * oldest pocl supported LLVM version.
+ *
+ */
+#define POCL_ADDRESS_SPACE_PRIVATE 0
+#define POCL_ADDRESS_SPACE_GLOBAL 1
+#define POCL_ADDRESS_SPACE_LOCAL 2
+#define POCL_ADDRESS_SPACE_CONSTANT 3
+#define POCL_ADDRESS_SPACE_GENERIC 4
 
-float4 _CL_OVERLOADABLE read_imagef (image2d_t image, sampler_t sampler,
-                                     int2 coord);
+#elif defined(__TCE__)
 
-float4 _CL_OVERLOADABLE read_imagef (image2d_t image, sampler_t sampler,
-                                     float2 coord);
+#define POCL_ADDRESS_SPACE_PRIVATE 0
+#define POCL_ADDRESS_SPACE_GLOBAL 3
+#define POCL_ADDRESS_SPACE_LOCAL 4
+#define POCL_ADDRESS_SPACE_CONSTANT 5
+#define POCL_ADDRESS_SPACE_GENERIC 6
 
-uint4 _CL_OVERLOADABLE read_imageui (image2d_t image, sampler_t sampler, 
-                                     int2 coord);
+#endif
 
-uint4 _CL_OVERLOADABLE read_imageui (image2d_t image, sampler_t sampler, 
-                                     int4 coord);
+typedef uint cl_mem_fence_flags;
 
-uint4 _CL_OVERLOADABLE read_imageui (image3d_t image, sampler_t sampler, 
-                                     int4 coord);
-
-int4 _CL_OVERLOADABLE read_imagei (image2d_t image, sampler_t sampler, 
-                                   int2 coord);
-
-
-void _CL_OVERLOADABLE write_imagei (image2d_t image, int2 coord, int4 color);
-
-void _CL_OVERLOADABLE write_imageui (image2d_t image, int2 coord, uint4 color);
-
-
-
-void _CL_OVERLOADABLE write_imagef (image2d_t image, int2 coord,
-                                    float4 color);
-/* not implemented 
-void _CL_OVERLOADABLE write_imagef (image2d_array_t image, int4 coord,
-                                    float4 color);
-
-void _CL_OVERLOADABLE write_imagei (image2d_array_t image, int4 coord,
-                                    int4 color);
-
-void _CL_OVERLOADABLE write_imageui (image2d_array_t image, int4 coord,
-                                     uint4 color);
-
-void _CL_OVERLOADABLE write_imagef (image1d_t image, int coord,
-                                    float4 color);
-
-void _CL_OVERLOADABLE write_imagei (image1d_t image, int coord,
-                                    int4 color);
-
-void _CL_OVERLOADABLE write_imageui (image1d_t image, int coord, 
-                                     uint4 color);
-
-void _CL_OVERLOADABLE write_imagef (image1d_buffer_t image, int coord, 
-                                    float4 color);
-
-void _CL_OVERLOADABLE write_imagei (image1d_buffer_t image, int coord,
-                                     int4 color);
-
-void _CL_OVERLOADABLE write_imageui (image1d_buffer_t image, int coord,
-                                     uint4 color);
-
-void _CL_OVERLOADABLE write_imagef (image1d_array_t image, int2 coord,
-                                    float4 color);
-
-void _CL_OVERLOADABLE write_imagei (image1d_array_t image, int2 coord,
-                                    int4 color);
-
-void _CL_OVERLOADABLE write_imageui (image1d_array_t image, int2 coord,
-                                     uint4 color);
-
-void _CL_OVERLOADABLE write_imageui (image3d_t image, int4 coord,
-                                     uint4 color);
-*/
-int _CL_OVERLOADABLE get_image_width (image1d_t image);
-int _CL_OVERLOADABLE get_image_width (image2d_t image);
-int _CL_OVERLOADABLE get_image_width (image3d_t image);
-
-int _CL_OVERLOADABLE get_image_height (image1d_t image);
-int _CL_OVERLOADABLE get_image_height (image2d_t image);
-int _CL_OVERLOADABLE get_image_height (image3d_t image);
-
-int _CL_OVERLOADABLE get_image_depth (image1d_t image);
-int _CL_OVERLOADABLE get_image_depth (image2d_t image);
-int _CL_OVERLOADABLE get_image_depth (image3d_t image);
-
-int2 _CL_OVERLOADABLE get_image_dim (image2d_t image);
-int2 _CL_OVERLOADABLE get_image_dim (image2d_array_t image);
-int4 _CL_OVERLOADABLE get_image_dim (image3d_t image);
+#include "_enable_all_exts.h"
 
 #endif
