@@ -38,6 +38,9 @@
 #include "pocl_device.h"
 #include "config.h"
 
+/* detects restrict, variadic macros etc */
+#include "pocl_compiler_features.h"
+
 #define POCL_FILENAME_LENGTH 1024
 
 typedef struct _mem_mapping mem_mapping_t;
@@ -306,21 +309,29 @@ struct _cl_command_node
 # define LLVM_OLDER_THAN_3_8 1
 # define LLVM_OLDER_THAN_3_9 1
 # define LLVM_OLDER_THAN_4_0 1
+# define LLVM_OLDER_THAN_5_0 1
 #endif
 
 #if (defined LLVM_3_7)
 # define LLVM_OLDER_THAN_3_8 1
 # define LLVM_OLDER_THAN_3_9 1
 # define LLVM_OLDER_THAN_4_0 1
+# define LLVM_OLDER_THAN_5_0 1
 #endif
 
 #if (defined LLVM_3_8)
 # define LLVM_OLDER_THAN_3_9 1
 # define LLVM_OLDER_THAN_4_0 1
+# define LLVM_OLDER_THAN_5_0 1
 #endif
 
 #if (defined LLVM_3_9)
 # define LLVM_OLDER_THAN_4_0 1
+# define LLVM_OLDER_THAN_5_0 1
+#endif
+
+#if (defined LLVM_4_0)
+# define LLVM_OLDER_THAN_5_0 1
 #endif
 
 #endif /* POCL_H */
