@@ -39,7 +39,7 @@ extern "C" {
 #pragma GCC visibility push(hidden)
 #endif
 
-void pocl_cache_init_topdir();
+int pocl_cache_init_topdir ();
 
 int
 pocl_cache_create_program_cachedir(cl_program program, unsigned device_i,
@@ -61,7 +61,9 @@ void pocl_cache_release_lock(void* lock);
 int pocl_cl_device_to_index(cl_program   program,
                             cl_device_id device);
 
-void pocl_cache_mk_temp_name(char* path);
+void pocl_cache_tempname (char *path_template, const char *suffix, int *fd);
+
+int pocl_cache_create_tempdir(char* path);
 
 int pocl_cache_write_program_source(char *program_cl_path,
                                     cl_program program);
