@@ -103,15 +103,15 @@ struct _cl_icd_dispatch pocl_dispatch = {
   &POclEnqueueWaitForEvents,
   &POclEnqueueBarrier,
   &POclGetExtensionFunctionAddress,
-  NULL, /* &POclCreateFromGLBuffer,      */
+  &POclCreateFromGLBuffer,
   &POclCreateFromGLTexture2D,
   &POclCreateFromGLTexture3D,
-  NULL, /* &POclCreateFromGLRenderbuffer, */
-  NULL, /* &POclGetGLObjectInfo,  70       */
-  NULL, /* &POclGetGLTextureInfo,        */
-  NULL, /* &POclEnqueueAcquireGLObjects, */
-  NULL, /* &POclEnqueueReleaseGLObjects, */
-  NULL, /* &POclGetGLContextInfoKHR,     */
+  &POclCreateFromGLRenderbuffer,
+  &POclGetGLObjectInfo,
+  &POclGetGLTextureInfo,
+  &POclEnqueueAcquireGLObjects,
+  &POclEnqueueReleaseGLObjects,
+  &POclGetGLContextInfoKHR,
   NULL, /* &clUnknown75 */
   NULL, /* &clUnknown76 */
   NULL, /* &clUnknown77 */
@@ -134,18 +134,18 @@ struct _cl_icd_dispatch pocl_dispatch = {
   &POclRetainDevice,
   &POclReleaseDevice,
   &POclCreateImage,
-  NULL, /* &POclCreateProgramWithBuiltInKernels, */
-  NULL, /* &POclCompileProgram,          */
-  NULL, /* &POclLinkProgram,             */
+  &POclCreateProgramWithBuiltInKernels,
+  &POclCompileProgram,
+  &POclLinkProgram,
   &POclUnloadPlatformCompiler, 
   &POclGetKernelArgInfo,
   &POclEnqueueFillBuffer,
   &POclEnqueueFillImage,
-  NULL, /* &POclEnqueueMigrateMemObjects, */
+  &POclEnqueueMigrateMemObjects,
   &POclEnqueueMarkerWithWaitList,
-  NULL, /* &POclEnqueueBarrierWithWaitList, */
-  NULL, /* &POclGetExtensionFunctionAddressForPlatform, */
-  NULL, /* &POclCreateFromGLTexture,     */
+  &POclEnqueueBarrierWithWaitList,
+  &POclGetExtensionFunctionAddressForPlatform,
+  &POclCreateFromGLTexture,
   NULL, /* &clUnknown109 */
   NULL, /* &clUnknown110 */
   NULL, /* &clUnknown111 */
@@ -221,9 +221,10 @@ struct _cl_icd_dispatch pocl_dispatch = {
 #endif
 };
 
-struct _cl_platform_id _platforms[1]  = {{&pocl_dispatch}};
+static struct _cl_platform_id _platforms[1]  = {{&pocl_dispatch}};
 #else
-struct _cl_platform_id _platforms[1]  = {};
+
+static struct _cl_platform_id _platforms[1] = {{ 1 }};
 #endif
 
 #ifdef __GNUC__
@@ -238,7 +239,7 @@ CL_API_ENTRY cl_int CL_API_CALL
 POname(clGetPlatformIDs)(cl_uint           num_entries,
                  cl_platform_id *  platforms,
                  cl_uint *         num_platforms) CL_API_SUFFIX__VERSION_1_0
-{	
+{
   const unsigned num = 1;
   unsigned i;
   
