@@ -43,9 +43,11 @@
 #  include "pocl_icd.h"
 #endif
 #include "pocl.h"
+#include "pocl_tracing.h"
 #include "pocl_debug.h"
 #include "pocl_hash.h"
-
+#include "pocl_runtime_config.h"
+#include "common.h"
 
 #if __STDC_VERSION__ < 199901L
 # if __GNUC__ >= 2
@@ -55,12 +57,6 @@
 # endif
 #endif
 
-#ifdef CLANGXX
-#define LINK_CMD CLANGXX
-#else
-#define LINK_CMD CLANG
-#endif
-
 typedef pthread_mutex_t pocl_lock_t;
 #define POCL_LOCK_INITIALIZER PTHREAD_MUTEX_INITIALIZER
 
@@ -70,23 +66,28 @@ typedef pthread_mutex_t pocl_lock_t;
 #define POCL_LOCK(__LOCK__) pthread_mutex_lock (&(__LOCK__))
 #define POCL_UNLOCK(__LOCK__) pthread_mutex_unlock (&(__LOCK__))
 #define POCL_INIT_LOCK(__LOCK__) pthread_mutex_init (&(__LOCK__), NULL)
-#define POCL_DESTROY_LOCK(__LOCK__) pthread_mutex_destroy (&(__LOCK__))
+/* We recycle OpenCL objects by not actually freeing them until the
+   very end. Thus, the lock should not be destoryed at the refcount 0. */
+#define POCL_DESTROY_LOCK(__LOCK__) pthread_mutex_destroy (&(__LOCK__)) 
 
 #define POCL_LOCK_OBJ(__OBJ__) POCL_LOCK((__OBJ__)->pocl_lock)
 #define POCL_UNLOCK_OBJ(__OBJ__) POCL_UNLOCK((__OBJ__)->pocl_lock)
 
-#define POCL_RELEASE_OBJECT(__OBJ__, __NEW_REFCOUNT__)  \
-  do {                                                  \
-    POCL_LOCK_OBJ (__OBJ__);                            \
-    __NEW_REFCOUNT__ = --(__OBJ__)->pocl_refcount;      \
-    POCL_UNLOCK_OBJ (__OBJ__);                          \
-    if (__NEW_REFCOUNT__ == 0) POCL_DESTROY_LOCK ((__OBJ__)->pocl_lock); \
+#define POCL_RELEASE_OBJECT(__OBJ__, __NEW_REFCOUNT__)                  \
+  do {                                                                  \
+    POCL_LOCK_OBJ (__OBJ__);                                            \
+    __NEW_REFCOUNT__ = --(__OBJ__)->pocl_refcount;                      \
+    assert((__OBJ__)->pocl_refcount >= 0);                              \
+    POCL_UNLOCK_OBJ (__OBJ__);                                          \
   } while (0)
 
+#define POCL_RETAIN_OBJECT_UNLOCKED(__OBJ__)    \
+    ++((__OBJ__)->pocl_refcount);
+
 #define POCL_RETAIN_OBJECT(__OBJ__)             \
   do {                                          \
     POCL_LOCK_OBJ (__OBJ__);                    \
-    (__OBJ__)->pocl_refcount++;                   \
+    POCL_RETAIN_OBJECT_UNLOCKED (__OBJ__);      \
     POCL_UNLOCK_OBJ (__OBJ__);                  \
   } while (0)
 
@@ -116,10 +117,17 @@ typedef pthread_mutex_t pocl_lock_t;
       POCL_INIT_OBJECT_NO_ICD(__OBJ__)
 #endif
 
+#define POCL_DESTROY_OBJECT(__OBJ__)                                          \
+  do                                                                          \
+    {                                                                         \
+      POCL_DESTROY_LOCK ((__OBJ__)->pocl_lock);                               \
+    }                                                                         \
+  while (0);
+
 /* Declares the generic pocl object attributes inside a struct. */
 #define POCL_OBJECT \
   pocl_lock_t pocl_lock; \
-  int pocl_refcount 
+  volatile int pocl_refcount 
 
 #define POCL_OBJECT_INIT \
   POCL_LOCK_INITIALIZER, 0
@@ -147,12 +155,12 @@ typedef pthread_mutex_t pocl_lock_t;
 /* Symbol aliases are supported */
 
 #  define POname(name) PO##name
-#  define POdeclsym(name)			\
+#  define POdeclsym(name)                      \
   __typeof__(name) PO##name __attribute__((visibility("hidden")));
 #  define POCL_ALIAS_OPENCL_SYMBOL(name)                                \
   __typeof__(name) name __attribute__((alias ("PO" #name), visibility("default")));
 #  define POsymAlways(name) POCL_ALIAS_OPENCL_SYMBOL(name)
-#  ifdef DIRECT_LINKAGE
+#  if !defined(BUILD_ICD)
 #    define POsym(name) POCL_ALIAS_OPENCL_SYMBOL(name)
 #  else
 #    define POsym(name)
@@ -164,10 +172,12 @@ typedef pthread_mutex_t pocl_lock_t;
  * it is used (as the ICD loader assumes that)*/
 #ifdef BUILD_ICD
 #  define POCL_ICD_OBJECT struct _cl_icd_dispatch *dispatch;
+#  define POCL_ICD_OBJECT_PLATFORM_ID POCL_ICD_OBJECT
 #  define POsymICD(name) POsym(name)
 #  define POdeclsymICD(name) POdeclsym(name)
 #else
 #  define POCL_ICD_OBJECT
+#  define POCL_ICD_OBJECT_PLATFORM_ID unsigned long id;
 #  define POsymICD(name)
 #  define POdeclsymICD(name)
 #endif
@@ -181,10 +191,13 @@ typedef pthread_mutex_t pocl_lock_t;
 #define POCL_HAS_KERNEL_ARG_TYPE_QUALIFIER     8
 #define POCL_HAS_KERNEL_ARG_NAME               16
 
-struct pocl_argument {
-  size_t size;
+typedef struct pocl_argument {
+  uint64_t size;
   void *value;
-};
+} pocl_argument;
+
+
+typedef struct event_node event_node;
 
 /**
  * Enumeration for kernel argument types
@@ -196,25 +209,82 @@ typedef enum {
   POCL_ARG_TYPE_SAMPLER = 3,
 } pocl_argument_type;
 
-struct pocl_argument_info {
+typedef struct pocl_argument_info {
+  char* type_name;
+  char* name;
+  cl_kernel_arg_address_qualifier address_qualifier;
+  cl_kernel_arg_access_qualifier access_qualifier;
+  cl_kernel_arg_type_qualifier type_qualifier;
   pocl_argument_type type;
   char is_local;
   char is_set;
-  cl_kernel_arg_address_qualifier address_qualifier;
-  cl_kernel_arg_access_qualifier access_qualifier;
-  char* type_name;
-  cl_kernel_arg_type_qualifier type_qualifier;
-  char* name;
-};
+  unsigned type_size;
+} pocl_argument_info;
 
 struct pocl_device_ops {
   const char *device_name;
-  void (*init_device_infos) (struct _cl_device_id*);
+  void *shared_data; /* data to be shared by a devices of same type */
+  void (*init_device_infos) (unsigned j, struct _cl_device_id*);
   /* implementation */
+
+  /* New driver api extension for out-of-order execution and
+     asynchronous devices.
+     See this for reference: http://URN.fi/URN:NBN:fi:tty-201412051583
+     See basic and pthread driver for reference. */
+
+  /* submit gives the command for the device. The command may be left in the cq
+     or stored to the device driver owning the cq. */
+  void (*submit) (_cl_command_node *node, cl_command_queue cq);
+
+  /* join is called by clFinish and this function blocks until all the enqueued
+     commands are finished. */
+  void (*join) (cl_device_id device, cl_command_queue cq);
+
+  /* flush is called when clFlush is called. This function ensures that
+     commands will be eventually executed. It is up to the device what happens
+     here, if anything. See basic and pthread for reference. */
+  void (*flush) (cl_device_id device, cl_command_queue cq);
+
+  /* notify is used to communicate to a device driver that an event, it has
+     been waiting, has been completed. */
+  void (*notify) (cl_device_id device, cl_event event, cl_event finished);
+
+  /* broadcast is(has to be) called by the device driver when a command is
+     completed.
+     It is used to broadcast notifications to device drivers waiting
+     this event to complete.
+     There is a default implementation for this. Use it if there is no need
+     to do anything special here. */
+  void (*broadcast) (cl_event event);
+
+  /* wait_event is blocking the execution until the waited event is complete.*/
+  void (*wait_event) (cl_device_id device, cl_event event);
+
+  /* update_event is an alternative way of handling event status changes if
+     something device spesific needs to be done when the status of the event
+     changes.
+     this function is called POCL_UPDATE_EVENT_* macros if available.
+     may be NULL, no need to implement if not needed. */
+  void (*update_event) (cl_device_id device, cl_event event, cl_int status);
+
+  /* free_event_data may be called when event is freed. Event data may only be
+     used by the device driver owning the corresponding command.
+     No need to implement this if the device does not need any event data. */
+  void (*free_event_data) (cl_event event);
+
+  /* /New driver api extension */
+
   void (*uninit) (cl_device_id device);
   unsigned int (*probe) (struct pocl_device_ops *ops);
-  void (*init) (cl_device_id device, const char *parameters);
-  cl_int (*alloc_mem_obj) (cl_device_id device, cl_mem mem_obj);
+  /* Device initialization. Parameters:
+   *  j : progressive index for the devices of the same type
+   *  device : struct to initialize
+   *  parameters : optional environment with device-specific parameters
+   */
+  cl_int (*init) (unsigned j, cl_device_id device, const char *parameters);
+  cl_int (*init_queue) (cl_command_queue queue);
+  void (*free_queue) (cl_command_queue queue);
+  cl_int (*alloc_mem_obj) (cl_device_id device, cl_mem mem_obj, void* host_ptr);
   void *(*create_sub_buffer) (void *data, void* buffer, size_t origin, size_t size);
   void (*free) (cl_device_id device, cl_mem mem_obj);
   void (*free_ptr) (cl_device_id device, void* mem_ptr);
@@ -249,7 +319,7 @@ struct pocl_device_ops {
                      size_t dst_row_pitch,
                      size_t dst_slice_pitch);
 
-void (*fill_rect) (void *data,
+  void (*fill_rect) (void *data,
                    void *__restrict__ const device_ptr,
                    const size_t *__restrict__ const buffer_origin,
                    const size_t *__restrict__ const region,
@@ -258,19 +328,19 @@ void (*fill_rect) (void *data,
                    void *fill_pixel,
                    size_t pixel_size);
 
-void (*memfill) (void *ptr,
-                 size_t size,
-                 size_t offset,
-                 const void* pattern,
-                 size_t pattern_size);
+  void (*memfill) (void *ptr,
+                   size_t size,
+                   size_t offset,
+                   const void* pattern,
+                   size_t pattern_size);
 
   /* Maps 'size' bytes of device global memory at buf_ptr + offset to 
      host-accessible memory. This might or might not involve copying 
      the block from the device. */
   void* (*map_mem) (void *data, void *buf_ptr, size_t offset, size_t size, void *host_ptr);
-  void* (*unmap_mem) (void *data, void *host_ptr, void *device_start_ptr, size_t size);
-  
-  void (*compile_submitted_kernels) (_cl_command_node* cmd);
+  void* (*unmap_mem) (void *data, void *host_ptr, void *device_start_ptr, size_t offset, size_t size);
+
+  void (*compile_kernel) (_cl_command_node* cmd, cl_kernel kernel, cl_device_id device);
   void (*run) (void *data, _cl_command_node* cmd);
   void (*run_native) (void *data, _cl_command_node* cmd);
 
@@ -281,7 +351,9 @@ void (*memfill) (void *ptr,
      owns the returned string. */
   char* (*init_build) (void *data);
 
-  void (*build_hash) (void *data, SHA1_CTX *build_hash);
+  void (*init_target_machine) (void *data, void *target_machine);
+
+  char* (*build_hash) (cl_device_id device);
 
     /* return supported image formats */
   cl_int (*get_supported_image_formats) (cl_mem_flags flags,
@@ -290,6 +362,7 @@ void (*memfill) (void *ptr,
 };
 
 typedef struct pocl_global_mem_t {
+  pocl_lock_t pocl_lock;
   size_t max_ever_allocated;
   size_t currently_allocated;
   size_t total_alloc_limit;
@@ -302,6 +375,10 @@ struct _cl_device_id {
   cl_device_type type;
   cl_uint vendor_id;
   cl_uint max_compute_units;
+  // for subdevices
+  cl_device_id parent_device;
+  unsigned core_start;
+  unsigned core_count;
   cl_uint max_work_item_dimensions;
   size_t max_work_item_sizes[3];
   size_t max_work_group_size;
@@ -326,6 +403,7 @@ struct _cl_device_id {
   cl_bool image_support;
   cl_uint max_read_image_args;
   cl_uint max_write_image_args;
+  cl_uint max_read_write_image_args;
   size_t image2d_max_width;
   size_t image2d_max_height;
   size_t image3d_max_width;
@@ -360,6 +438,11 @@ struct _cl_device_id {
      we need to generate work-item loops to execute all the work-items
      in the WG, otherwise the hardware spawns the WIs. */
   cl_bool spmd;
+  /* The Workgroup pass creates launcher functions and replaces work-item
+     placeholder global variables (e.g. _local_size_, _global_offset_ etc) with
+     loads from the context struct passed as a kernel argument. This flag
+     enables or disables this pass. */
+  cl_bool workgroup_pass;
   cl_device_exec_capabilities execution_capabilities;
   cl_command_queue_properties queue_properties;
   cl_platform_id platform;
@@ -372,7 +455,6 @@ struct _cl_device_id {
   char *short_name;
   char *long_name;
   char *cache_dir_name;
-  cl_device_id parent_device;
 
   const char *vendor;
   const char *driver_version;
@@ -383,12 +465,13 @@ struct _cl_device_id {
   cl_uint cl_version_minor;    // 0
   const char *cl_version_std;  // "CL2.0"
   cl_uint cl_version_int;      // 200
- 
+
   void *data;
   const char* llvm_target_triplet; /* the llvm target triplet to use */
   const char* llvm_cpu; /* the llvm CPU variant to use */
-  /* A running number (starting from zero) across all the device instances. Used for 
-     indexing  arrays in data structures with device specific entries. */
+  /* A running number (starting from zero) across all the device instances.
+     Used for indexing arrays in data structures with device specific
+     entries. */
   int dev_id;
   int global_mem_id; /* identifier for device global memory */
   /* pointer to an accounting struct for global memory */
@@ -397,8 +480,13 @@ struct _cl_device_id {
   /* Convert automatic local variables to kernel arguments? */
   int autolocals_to_args;
 
-  // True if the device supports SVM has priority
-  // at allocating Shared Virtual Memory
+  /* The target specific IDs for the different OpenCL address spaces. */
+  unsigned global_as_id;
+  unsigned local_as_id;
+  unsigned constant_as_id;
+
+  /* True if the device supports SVM. Then it has the responsibility of
+     allocating shared buffers residing in Shared Virtual Memory areas. */
   cl_bool should_allocate_svm;
   /* OpenCL 2.0 properties */
   cl_device_svm_capabilities svm_caps;
@@ -413,6 +501,7 @@ struct _cl_device_id {
   cl_command_queue_properties on_host_queue_props;
 
   struct pocl_device_ops *ops; /* Device operations, shared amongst same devices */
+
 };
 
 #define DEVICE_SVM_FINEGR(dev) (dev->svm_caps & (CL_DEVICE_SVM_FINE_GRAIN_BUFFER \
@@ -423,8 +512,12 @@ struct _cl_device_id {
 
 #define DEVICE_MMAP_IS_NOP(dev) (DEVICE_SVM_FINEGR(dev) && DEVICE_SVM_ATOM(dev))
 
+#define CHECK_DEVICE_AVAIL_RET(dev) if(!dev->available) { POCL_MSG_ERR("This cl_device is not available.\n"); return CL_INVALID_DEVICE; }
+#define CHECK_DEVICE_AVAIL_RETV(dev) if(!dev->available) { POCL_MSG_ERR("This cl_device is not available.\n"); return; }
+
+
 struct _cl_platform_id {
-  POCL_ICD_OBJECT
+  POCL_ICD_OBJECT_PLATFORM_ID
 }; 
 
 struct _cl_context {
@@ -450,6 +543,14 @@ struct _cl_context {
 
 };
 
+typedef struct _pocl_data_sync_item pocl_data_sync_item;
+struct _pocl_data_sync_item {
+  unsigned int volatile event_id;
+  cl_event volatile event;
+  pocl_data_sync_item *volatile next;
+};
+
+struct _cl_event;
 struct _cl_command_queue {
   POCL_ICD_OBJECT
   POCL_OBJECT;
@@ -458,12 +559,20 @@ struct _cl_command_queue {
   cl_device_id device;
   cl_command_queue_properties properties;
   /* implementation */
-  _cl_command_node *root;
+  cl_event events; /* events of the enqueued commands in enqueue order */
+  struct _cl_event * volatile barrier;
+  volatile int command_count; /* counter for unfinished command enqueued */
+  volatile pocl_data_sync_item last_event;
+
+  /* backend specific data */
+  void *data;
 };
 
+
 /* memory identifier: id to point the global memory where memory resides 
                       + pointer to actual data */
-typedef struct _pocl_mem_identifier{
+typedef struct _pocl_mem_identifier {
+  int available; /* ... in this mem objs context */
   int global_mem_id;
   void* mem_ptr;
 } pocl_mem_identifier;
@@ -476,6 +585,7 @@ struct _cl_mem {
   cl_mem_object_type type;
   cl_mem_flags flags;
   size_t size;
+  size_t origin; /* for sub-buffers */
   void *mem_host_ptr;
   cl_uint map_count;
   cl_context context;
@@ -489,12 +599,22 @@ struct _cl_mem {
      The location of the device's buffer ptr is determined by
      the device's dev_id. */
   pocl_mem_identifier *device_ptrs;
+  /* device that allocated, and is going to free, 
+     the shared system mem allocation */
+  cl_device_id shared_mem_allocation_owner;
+  /* device where this mem obj resides */
+  volatile cl_device_id owning_device;
+  /* latest event assosiated with the buffer, set NULL when event completed */
+  volatile cl_event latest_event;
   /* A linked list of regions of the buffer mapped to the 
      host memory */
   mem_mapping_t *mappings;
   /* in case this is a sub buffer, this points to the parent
      buffer */
   cl_mem_t *parent;
+  /* A linked list of destructor callbacks */
+  mem_destructor_callback_t *destructor_callbacks;
+
   /* Image flags */
   cl_bool                 is_image;
   cl_channel_order        image_channel_order;
@@ -510,9 +630,11 @@ struct _cl_mem {
   cl_uint                 num_mip_levels;
   cl_uint                 num_samples;
   cl_mem                  buffer;
-  /* Pipe specific */
-  cl_uint packet_size;
-  cl_uint max_packets;
+
+  /* pipe flags */
+  cl_bool                 is_pipe;
+  size_t                  pipe_packet_size;
+  size_t                  pipe_max_packets;
 };
 
 typedef uint8_t SHA1_digest_t[SHA1_DIGEST_SIZE * 2 + 1];
@@ -523,15 +645,31 @@ struct _cl_program {
   /* queries */
   cl_context context;
   cl_uint num_devices;
+  /* bool flag, set to 1 when removing/adding default kernels to a program
+   * This code needs to be eventually fixed by introducing kernel_metadata
+   * struct, see Issue #390 */
+  int operating_on_default_kernels;
+  /* -cl-denorms-are-zero build option */
+  unsigned flush_denorms;
   cl_device_id *devices;
   /* all the program sources appended together, terminated with a zero */
   char *source;
   /* The options in the last clBuildProgram call for this Program. */
   char *compiler_options;
-  /* The binaries for each device. Currently the binary is directly the
-     sequential bitcode produced from the kernel sources.*/
-  size_t *binary_sizes; 
-  unsigned char **binaries; 
+  /* The binaries for each device.  Currently the binary is directly the
+     sequential bitcode produced from the kernel sources.  */
+  size_t *binary_sizes;
+  unsigned char **binaries;
+
+  /* Poclcc binary format.  */
+  size_t *pocl_binary_sizes;
+  unsigned char **pocl_binaries;
+
+  /* "Default" kernels. See: https://github.com/pocl/pocl/issues/390  */
+  size_t num_kernels;
+  char **kernel_names;
+  cl_kernel *default_kernels;
+
   /* implementation */
   cl_kernel kernels;
   /* Per-device program hash after build */
@@ -546,6 +684,8 @@ struct _cl_program {
   void** read_locks;
   /* Use to store build status */
   cl_build_status build_status;
+  /* Use to store binary type */
+  cl_program_binary_type binary_type;
 };
 
 struct _cl_kernel {
@@ -556,16 +696,18 @@ struct _cl_kernel {
   cl_uint num_args;
   cl_context context;
   cl_program program;
-  /* implementation */
-  lt_dlhandle dlhandle;
   struct pocl_argument_info *arg_info;
   cl_bitfield has_arg_metadata;
+  char *attributes;
   cl_uint num_locals;
-  int *reqd_wg_size;
+  size_t *reqd_wg_size;
   /* The kernel arguments that are set with clSetKernelArg().
      These are copied to the command queue command at enqueue. */
   struct pocl_argument *dyn_arguments;
   struct _cl_kernel *next;
+
+  /* backend specific data */
+  void *data;
 };
 
 typedef struct event_callback_item event_callback_item;
@@ -577,6 +719,13 @@ struct event_callback_item
   struct event_callback_item *next;
 };
 
+
+struct event_node
+{
+  cl_event event;
+  event_node * volatile next;
+};
+
 typedef struct _cl_event _cl_event;
 struct _cl_event {
   POCL_ICD_OBJECT
@@ -584,85 +733,196 @@ struct _cl_event {
   cl_context context;
   cl_command_queue queue;
   cl_command_type command_type;
+  _cl_command_node *command;
+  unsigned int id;
 
   /* list of callback functions */
-  event_callback_item* callback_list;
+  event_callback_item *volatile callback_list;
+
+  /* list of devices needing completion notification for this event */
+  event_node * volatile notify_list;
+  event_node * volatile wait_list;
+
+  /* OoO doesn't use sync points -> put used buffers here */
+  cl_mem *mem_objs;
+  int num_buffers;
 
   /* The execution status of the command this event is monitoring. */
-  cl_int status;
+  volatile cl_int status;
 
   /* Profiling data: time stamps of the different phases of execution. */
   cl_ulong time_queue;  /* the enqueue time */
   cl_ulong time_submit; /* the time the command was submitted to the device */
   cl_ulong time_start;  /* the time the command actually started executing */
-  cl_ulong time_end;    /* the finish time of the command */   
+  cl_ulong time_end;    /* the finish time of the command */
+
+  void *data; /* Device specific data */  
 
   /* impicit event = an event for pocl's internal use, not visible to user */
   int implicit_event;
   _cl_event * volatile next;
+  _cl_event * volatile prev;
 };
 
-typedef struct _cl_sampler cl_sampler_t;
+typedef struct _pocl_user_event_data
+{
+  pthread_cond_t wakeup_cond;
+  pthread_mutex_t lock;
+} pocl_user_event_data;
 
+typedef struct _cl_sampler cl_sampler_t;
 struct _cl_sampler {
   POCL_ICD_OBJECT
+  POCL_OBJECT;
+  cl_context context;
   cl_bool             normalized_coords;
   cl_addressing_mode  addressing_mode;
   cl_filter_mode      filter_mode;
 };
 
-#define POCL_UPDATE_EVENT_QUEUED(__event)                               \
-  do {                                                                  \
-    if ((__event) != NULL && (*(__event)) != NULL)                      \
-      {                                                                 \
-        cl_command_queue __cq = (*(__event))->queue;                    \
-        (*(__event))->status = CL_QUEUED;                               \
-        if (__cq && __cq->properties & CL_QUEUE_PROFILING_ENABLE)       \
-          (*(__event))->time_queue =                                    \
-            __cq->device->ops->get_timer_value(__cq->device->data);     \
-      }                                                                 \
-  } while (0)                                                           \
+#define POCL_UPDATE_EVENT_QUEUED(__event)                                     \
+  do                                                                          \
+    {                                                                         \
+      if ((__event) != NULL)                                                  \
+        {                                                                     \
+          cl_command_queue __cq = (__event)->queue;                           \
+          if ((__cq)->device->ops->update_event)                              \
+            (__cq)->device->ops->update_event ((__cq)->device, (__event),     \
+                                               CL_QUEUED);                    \
+          else                                                                \
+            {                                                                 \
+              (__event)->status = CL_QUEUED;                                  \
+              if (__cq && __cq->properties & CL_QUEUE_PROFILING_ENABLE)       \
+                (__event)->time_queue = __cq->device->ops->get_timer_value (  \
+                    __cq->device->data);                                      \
+            }                                                                 \
+          pocl_event_updated (__event, CL_QUEUED);                            \
+        }                                                                     \
+    }                                                                         \
+  while (0)
 
-#define POCL_UPDATE_EVENT_SUBMITTED(__event)                            \
-  do {                                                                  \
-    if ((__event) != NULL && (*(__event)) != NULL)                      \
-      {                                                                 \
-        assert((*(__event))->status == CL_QUEUED);                      \
-        (*(__event))->status = CL_SUBMITTED;                            \
-        cl_command_queue __cq = (*(__event))->queue;                    \
-        if (__cq && __cq->properties & CL_QUEUE_PROFILING_ENABLE)       \
-          (*(__event))->time_submit =                                   \
-            __cq->device->ops->get_timer_value(__cq->device->data);     \
-      }                                                                 \
-  } while (0)                                                           \
+#define POCL_UPDATE_EVENT_SUBMITTED(__event)                                  \
+  do                                                                          \
+    {                                                                         \
+      if ((__event) != NULL)                                                  \
+        {                                                                     \
+          assert ((__event)->status == CL_QUEUED);                            \
+          cl_command_queue __cq = (__event)->queue;                           \
+          if ((__cq)->device->ops->update_event)                              \
+            (__cq)->device->ops->update_event ((__cq)->device, (__event),     \
+                                               CL_SUBMITTED);                 \
+          else                                                                \
+            {                                                                 \
+              (__event)->status = CL_SUBMITTED;                               \
+              if (__cq && __cq->properties & CL_QUEUE_PROFILING_ENABLE)       \
+                (__event)->time_submit = __cq->device->ops->get_timer_value ( \
+                    __cq->device->data);                                      \
+            }                                                                 \
+          pocl_event_updated (__event, CL_SUBMITTED);                         \
+        }                                                                     \
+    }                                                                         \
+  while (0)
 
-#define POCL_UPDATE_EVENT_RUNNING(__event)                              \
-  do {                                                                  \
-    if (__event != NULL && (*(__event)) != NULL)                        \
-      {                                                                 \
-        assert((*(__event))->status == CL_SUBMITTED);                   \
-        (*(__event))->status = CL_RUNNING;                              \
-        cl_command_queue __cq = (*(__event))->queue;                    \
-        if (__cq && __cq->properties & CL_QUEUE_PROFILING_ENABLE)       \
-          (*(__event))->time_start =                                    \
-            __cq->device->ops->get_timer_value(__cq->device->data);     \
-      }                                                                 \
-  } while (0)                                                           \
+#define POCL_UPDATE_EVENT_RUNNING(__event)                                    \
+  do                                                                          \
+    {                                                                         \
+      if (__event != NULL)                                                    \
+        {                                                                     \
+          assert ((__event)->status == CL_SUBMITTED);                         \
+          cl_command_queue __cq = (__event)->queue;                           \
+          if ((__cq)->device->ops->update_event)                              \
+            (__cq)->device->ops->update_event ((__cq)->device, (__event),     \
+                                               CL_RUNNING);                   \
+          else                                                                \
+            {                                                                 \
+              (__event)->status = CL_RUNNING;                                 \
+              if (__cq && __cq->properties & CL_QUEUE_PROFILING_ENABLE)       \
+                (__event)->time_start = __cq->device->ops->get_timer_value (  \
+                    __cq->device->data);                                      \
+            }                                                                 \
+          pocl_event_updated (__event, CL_RUNNING);                           \
+        }                                                                     \
+    }                                                                         \
+  while (0)
 
-#define POCL_UPDATE_EVENT_COMPLETE(__event)                             \
-  do {                                                                  \
-    if ((__event) != NULL && (*(__event)) != NULL)                      \
-      {                                                                 \
-        assert((*(__event))->status == CL_RUNNING);                     \
-        (*(__event))->status = CL_COMPLETE;                             \
-        cl_command_queue __cq = (*(__event))->queue;                    \
-        if (__cq && __cq->properties & CL_QUEUE_PROFILING_ENABLE)       \
-          (*(__event))->time_end =                                      \
-            __cq->device->ops->get_timer_value(__cq->device->data);     \
-      }                                                                 \
-  } while (0)                                                           \
+#define POCL_UPDATE_EVENT_COMPLETE_INNER(__event, POST_EVENT)                 \
+  do                                                                          \
+    {                                                                         \
+      if ((__event) != NULL)                                                  \
+        {                                                                     \
+          assert ((__event)->status == CL_RUNNING);                           \
+          cl_command_queue __cq = (__event)->queue;                           \
+          assert ((__event)->status == CL_RUNNING);                           \
+          if ((__cq)->device->ops->update_event)                              \
+            (__cq)->device->ops->update_event ((__cq)->device, (__event),     \
+                                               CL_COMPLETE);                  \
+          else                                                                \
+            {                                                                 \
+              pocl_mem_objs_cleanup (__event);                                \
+              POCL_LOCK_OBJ (__event);                                        \
+              (__event)->status = CL_COMPLETE;                                \
+              if ((__cq)->properties & CL_QUEUE_PROFILING_ENABLE)             \
+                {                                                             \
+                  (__event)->time_end                                         \
+                      = (__cq)->device->ops->get_timer_value (                \
+                          (__cq)->device->data);                              \
+                }                                                             \
+              POCL_UNLOCK_OBJ (__event);                                      \
+              (__cq)->device->ops->broadcast (__event);                       \
+              pocl_update_command_queue (__event);                            \
+            }                                                                 \
+          pocl_event_updated (__event, CL_COMPLETE);                          \
+          POST_EVENT;                                                         \
+          POname (clReleaseEvent) (__event);                                  \
+        }                                                                     \
+    }                                                                         \
+  while (0)
+
+#define POCL_UPDATE_EVENT_COMPLETE(__event)                                   \
+  POCL_UPDATE_EVENT_COMPLETE_INNER (__event, NULL)
+
+#define POCL_UPDATE_EVENT_COMPLETE_MSG(__event, msg)                          \
+  POCL_UPDATE_EVENT_COMPLETE_INNER (__event,                                  \
+                                    POCL_DEBUG_EVENT_TIME ((__event), msg))
+
+#define CL_FAILED (-1)
+
+#define POCL_UPDATE_EVENT_FAILED(__event)                                     \
+  do                                                                          \
+    {                                                                         \
+      if ((__event) != NULL)                                                  \
+        {                                                                     \
+          cl_command_queue __cq = (__event)->queue;                           \
+          if ((__cq)->device->ops->update_event)                              \
+            (__cq)->device->ops->update_event ((__cq)->device, (__event),     \
+                                               CL_FAILED);                    \
+          else                                                                \
+            {                                                                 \
+              pocl_mem_objs_cleanup (__event);                                \
+              POCL_LOCK_OBJ (__event);                                        \
+              if ((__event)->status > CL_COMPLETE)                            \
+                (__event)->status = CL_FAILED;                                \
+              if ((__cq)->properties & CL_QUEUE_PROFILING_ENABLE)             \
+                {                                                             \
+                  (__event)->time_end                                         \
+                      = (__cq)->device->ops->get_timer_value (                \
+                          (__cq)->device->data);                              \
+                }                                                             \
+              POCL_UNLOCK_OBJ (__event);                                      \
+              (__cq)->device->ops->broadcast (__event);                       \
+              pocl_update_command_queue (__event);                            \
+            }                                                                 \
+          pocl_event_updated (__event, CL_FAILED);                            \
+          POname (clReleaseEvent) (__event);                                  \
+        }                                                                     \
+    }                                                                         \
+  while (0)
+
+#ifndef __cplusplus
 
 #define min(a,b) (((a) < (b)) ? (a) : (b))
 #define max(a,b) (((a) > (b)) ? (a) : (b))
 
+#endif
+
 #endif /* POCL_CL_H */
