@@ -22,19 +22,28 @@
    THE SOFTWARE.
 */
 
+#pragma clang diagnostic push
+#pragma clang diagnostic ignored "-Wformat-security"
+
 #include <limits.h>
 #include <stdarg.h>
 #include <stdbool.h>
+#include <math.h>
 
 // We implement the OpenCL printf by calling the C99 printf. This is
 // not very efficient, but is easy to implement.
+#if LLVM_OLDER_THAN_5_0
 #define OCL_C_AS __attribute__((address_space(0)))
+#else
+#define OCL_C_AS
+#endif
 int printf(OCL_C_AS const char* restrict fmt, ...);
-int snprintf(OCL_C_AS char* restrict str, size_t size, OCL_C_AS const char* restrict fmt, ...);
+int snprintf(OCL_C_AS char* restrict str, size_t size,
+             OCL_C_AS const char* restrict fmt, ...);
 
 // For debugging
 // Use as: DEBUG_PRINTF((fmt, args...)) -- note double parentheses!
-// #define DEBUG_PRINTF(args) (printf args)
+//#define DEBUG_PRINTF(args) (printf args)
 #define DEBUG_PRINTF(args) ((void)0)
 
 // Conversion flags
@@ -57,12 +66,13 @@ typedef struct {
 
 #define DEFINE_PRINT_INTS(WIDTH)                                        \
   void _cl_print_ints_##WIDTH(flags_t flags, int field_width, int precision, \
-                              char conv, OCL_C_AS const void* vals, int n)       \
+                              char conv, OCL_C_AS const void* vals, int n) \
   {                                                                     \
     DEBUG_PRINTF(("[printf:ints:n=%df]\n", n));                         \
     char outfmt[1000];                                                  \
+    OCL_C_AS char str[] = "%%%s%s%s%s%s%.0d%s%.0d" INT_CONV_##WIDTH "%c"; \
     snprintf(outfmt, sizeof outfmt,                                     \
-             "%%%s%s%s%s%s%.0d%s%.0d" INT_CONV_##WIDTH "%c",            \
+             str,                                                       \
              flags.left ? "-" : "",                                     \
              flags.plus ? "+" : "",                                     \
              flags.space ? " " : "",                                    \
@@ -73,10 +83,11 @@ typedef struct {
              precision != -1 ? precision : 0,                           \
              conv);                                                     \
     DEBUG_PRINTF(("[printf:ints:outfmt=%s]\n", outfmt));                \
+    OCL_C_AS char comma[] = ",";                                        \
     for (int d=0; d<n; ++d) {                                           \
       DEBUG_PRINTF(("[printf:ints:d=%d]\n", d));                        \
-      if (d != 0) printf(",");                                          \
-      printf(outfmt, ((OCL_C_AS const WIDTH*)vals)[d]);                          \
+      if (d != 0) printf(comma);                                        \
+      printf(outfmt, ((OCL_C_AS const WIDTH*)vals)[d]);                 \
     }                                                                   \
     DEBUG_PRINTF(("[printf:ints:done]\n"));                             \
   }
@@ -95,13 +106,11 @@ DEFINE_PRINT_INTS(long)
 // Helper routines to output floats
 
 // Defined in OpenCL
-float __attribute__((overloadable)) vload_half(size_t offset, const half *p);
+float __attribute__((overloadable)) vload_half(size_t offset,
+                                               OCL_C_AS const half *p);
 
 // Note: To simplify implementation, we print double values with %lf,
 // although %f would suffice as well
-#define FLOAT_CONV_half   "h"
-#define FLOAT_CONV_float  ""
-#define FLOAT_CONV_double "l"
 #define FLOAT_GET_half(ptr)   vload_half(0, ptr)
 #define FLOAT_GET_float(ptr)  (*(ptr))
 #define FLOAT_GET_double(ptr) (*(ptr))
@@ -112,8 +121,9 @@ float __attribute__((overloadable)) vload_half(size_t offset, const half *p);
   {                                                                     \
     DEBUG_PRINTF(("[printf:floats:n=%dd]\n", n));                       \
     char outfmt[1000];                                                  \
+    OCL_C_AS char str[] = "%%%s%s%s%s%s%.0d%s%.0d" "%c";                \
     snprintf(outfmt, sizeof outfmt,                                     \
-             "%%%s%s%s%s%s%.0d%s%.0d" FLOAT_CONV_##WIDTH "%c",          \
+             str,                                                       \
              flags.left ? "-" : "",                                     \
              flags.plus ? "+" : "",                                     \
              flags.space ? " " : "",                                    \
@@ -124,10 +134,14 @@ float __attribute__((overloadable)) vload_half(size_t offset, const half *p);
              precision != -1 ? precision : 0,                           \
              conv);                                                     \
     DEBUG_PRINTF(("[printf:floats:outfmt=%s]\n", outfmt));              \
+    OCL_C_AS char comma[] = ",";                                        \
     for (int d=0; d<n; ++d) {                                           \
       DEBUG_PRINTF(("[printf:floats:d=%d]\n", d));                      \
-      if (d != 0) printf(",");                                          \
-      printf(outfmt, FLOAT_GET_##WIDTH((OCL_C_AS const WIDTH*)vals+d));          \
+      if (d != 0) printf(comma);                                        \
+      WIDTH val = (FLOAT_GET_##WIDTH((OCL_C_AS const WIDTH*)vals+d));   \
+      if (isnan (val))                                                  \
+        val = NAN;                                                      \
+      printf(outfmt, (double)val);                                      \
     }                                                                   \
     DEBUG_PRINTF(("[printf:floats:done]\n"));                           \
   }
@@ -150,8 +164,9 @@ void _cl_print_char(flags_t flags, int field_width, int val)
 {
   DEBUG_PRINTF(("[printf:char]\n"));
   char outfmt[1000];
+  char string[] = "%%%s%.0dc";
   snprintf(outfmt, sizeof outfmt,
-           "%%%s%.0dc",
+           string,
            flags.left ? "-" : "",
            field_width);
   DEBUG_PRINTF(("[printf:char:outfmt=%s]\n", outfmt));
@@ -159,14 +174,19 @@ void _cl_print_char(flags_t flags, int field_width, int val)
   DEBUG_PRINTF(("[printf:char:done]\n"));
 }
 
-void _cl_print_string(flags_t flags, int field_width, OCL_C_AS const char* val)
+void
+_cl_print_string (flags_t flags, int field_width, int precision,
+                  OCL_C_AS const char *val)
 {
   DEBUG_PRINTF(("[printf:char]\n"));
   char outfmt[1000];
+  char string[] = "%%%s%.0d%s%.0ds";
   snprintf(outfmt, sizeof outfmt,
-           "%%%s%.0ds",
+           string,
            flags.left ? "-" : "",
-           field_width);
+           field_width,
+           (precision > 0) ? "." : "",
+           (precision > 0) ? precision : 0);
   DEBUG_PRINTF(("[printf:char:outfmt=%s]\n", outfmt));
   printf(outfmt, val);
   DEBUG_PRINTF(("[printf:char:done]\n"));
@@ -176,8 +196,9 @@ void _cl_print_pointer(flags_t flags, int field_width, OCL_C_AS const void* val)
 {
   DEBUG_PRINTF(("[printf:char]\n"));
   char outfmt[1000];
+  char string[] = "%%%s%.0dp";
   snprintf(outfmt, sizeof outfmt,
-           "%%%s%.0dp",
+           string,
            flags.left ? "-" : "",
            field_width);
   DEBUG_PRINTF(("[printf:char:outfmt=%s]\n", outfmt));
@@ -199,8 +220,16 @@ void _cl_print_pointer(flags_t flags, int field_width, OCL_C_AS const void* val)
 // - if there is an error during parsing, a "goto error" aborts the
 //   routine, returning -1
 
+// This should be queried from the machine in the non-TAS case.
+// For now assume that if we are not using the fake address space
+// ids then we have a single address space. This version of printf
+// doesn't work with multiple address spaces anyways.
+#ifdef POCL_USE_FAKE_ADDR_SPACE_IDS
 #define OCL_CONSTANT_AS __attribute__((address_space(3)))
-int _cl_printf(const OCL_CONSTANT_AS char* restrict format, ...)
+#else
+#define OCL_CONSTANT_AS
+#endif
+int __cl_printf(const OCL_CONSTANT_AS char* restrict format, ...)
 {
   DEBUG_PRINTF(("[printf:format=%s]\n", format));
   va_list ap;
@@ -213,7 +242,8 @@ int _cl_printf(const OCL_CONSTANT_AS char* restrict format, ...)
       
       if (ch == '%') {
         DEBUG_PRINTF(("[printf:%%]\n"));
-        printf("%%");           // literal %
+        char s[] = "%%";
+        printf(s);           // literal %
         ch = *++format;
       } else {
         DEBUG_PRINTF(("[printf:arg]\n"));
@@ -412,11 +442,10 @@ int _cl_printf(const OCL_CONSTANT_AS char* restrict format, ...)
           // Output a string
         case 's': {
           if (flags.plus || flags.space || flags.alt || flags.zero) goto error;
-          if (precision != -1) goto error;
           if (vector_length != 1) goto error;
           if (length != 0) goto error;
           OCL_C_AS const char* val = va_arg(ap, OCL_C_AS const char*);
-          _cl_print_string(flags, field_width, val);
+          _cl_print_string (flags, field_width, precision, val);
           break;
         }
           
@@ -439,7 +468,8 @@ int _cl_printf(const OCL_CONSTANT_AS char* restrict format, ...)
 
     } else {
       DEBUG_PRINTF(("[printf:literal]\n"));
-      printf("%c", ch);
+      char literal[] = "%c";
+      printf(literal, ch);
       ch = *++format;
     }
   }
@@ -451,6 +481,9 @@ int _cl_printf(const OCL_CONSTANT_AS char* restrict format, ...)
  error:;
   va_end(ap);
   DEBUG_PRINTF(("[printf:error]\n"));
-  printf("(printf format string error)");
+  char string [] = "(printf format string error)";
+  printf(string);
   return -1;
 }
+
+#pragma clang diagnostic pop
